#include "Cv7_main.h"

#include "glm/gtx/color_space.hpp"

//---------------------
//----    SCENE    ----
//---------------------

/// Reloads all shaders
void reload_shaders()
{
	// Shader program that is used when when the particles are rendered as points
	particle_simple_program.Init();
	particle_simple_program.AddVertexShader("Shaders/particle_simple_vertex.glsl");
	particle_simple_program.AddFragmentShader("Shaders/particle_simple_fragment.glsl");
	particle_simple_program.Link();

	// Shader program that is used when when the particles are rendered as quads
	particle_textured_program.Init();
	particle_textured_program.AddVertexShader("Shaders/particle_textured_vertex.glsl");
	particle_textured_program.AddGeometryShader("Shaders/particle_textured_geometry.glsl");
	particle_textured_program.AddFragmentShader("Shaders/particle_textured_fragment.glsl");
	particle_textured_program.Link();

	// Compute shader that updates the position of particles
	update_particles_program.Init();
	update_particles_program.AddComputeShader("Shaders/nbody_compute.glsl");
	update_particles_program.Link();

	// Compute shader that updates the position of particles that uses shared memory to speed up the computation
	update_particles_shared_program.Init();
	update_particles_shared_program.AddComputeShader("Shaders/nbody_shared_compute.glsl");
	update_particles_shared_program.Link();

	cout << "Shaders are reloaded" << endl;
}

void reset_particles()
{
	// Reinitializes the positions and velocities of all particles in the simulation. The points
	// are uniformly distributed on the surface of a unit sphere, and their initial velocity is zero.

	// Set seed. Set it here to ensure that though all particles are at random places, they are always
	// at the same places when they are reset.
	srand(reset_particles_seed);

	for (int i = 0; i < current_particle_count; i++)
	{
		float alpha = float(rand()) / float(RAND_MAX) * 2.0f * float(M_PI);
		float beta = asinf(float(rand()) / float(RAND_MAX) * 2.0f - 1.0f);
		glm::vec3 point_on_sphere = glm::vec3(cosf(alpha) * cosf(beta), sinf(alpha) * cosf(beta), sinf(beta));

		particle_positions[0][i] = glm::vec4(point_on_sphere, 1.0f);
		particle_positions[1][i] = glm::vec4(point_on_sphere, 1.0f);
		particle_velocities[i] = glm::vec4(0.0f);
	}

	// Update the OpenGL buffers accordingly
	glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[0]);
	glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(glm::vec4) * current_particle_count, particle_positions[0].data());
	glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[1]);
	glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(glm::vec4) * current_particle_count, particle_positions[1].data());
	glBindBuffer(GL_ARRAY_BUFFER, particle_velocities_bo);
	glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(glm::vec4) * current_particle_count, particle_velocities.data());
	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	PrintOpenGLInfo();

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare all cameras

	the_camera.SetCamera(-0.5f, -0.5f, 15.0f);
	CameraData_ubo.Init();

	//----------------------------------------------
	//--  Prepare textures

	particle_tex = CreateAndLoadTexture2D(L"../../textures/star.png");
	glBindTexture(GL_TEXTURE_2D, particle_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);
	glGenerateMipmap(GL_TEXTURE_2D);		// Particles are really small, use mipmaps for them
	glBindTexture(GL_TEXTURE_2D, 0);
	
	//----------------------------------------------
	//--  Prepare geometries

	// Create CPU buffers with the data of particles - positions and velocities
	particle_positions[0].resize(max_particle_count, glm::vec4(0.0f));
	particle_positions[1].resize(max_particle_count, glm::vec4(0.0f));
	particle_velocities.resize(max_particle_count, glm::vec4(0.0f));

	// Compute random color for each particle
	srand(12345);
	vector<glm::vec3> particle_colors(max_particle_count);
	for (int i = 0; i < max_particle_count; i++)
	{
		// Give the particle a random color. Thanks to the fact that GLM has functions for HSV -> RGB conversion,
		// we choose hue randomly, use 1.0 for saturation and value, and convert it to RGB, thus obtaining
		// a uniformly random color.
		particle_colors[i] = glm::rgbColor(glm::vec3(float(rand()) / float(RAND_MAX) * 360.0f, 1.0f, 1.0f));
	}

	// Allocate GPU buffers, update the data of the colors
	glGenBuffers(2, particle_positions_bo);
	glGenBuffers(1, &particle_velocities_bo);
	glGenBuffers(1, &particle_colors_bo);
	glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(float) * 4 * max_particle_count, nullptr, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(float) * 4 * max_particle_count, nullptr, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, particle_velocities_bo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(float) * 4 * max_particle_count, nullptr, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, particle_colors_bo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(float) * 3 * max_particle_count, particle_colors.data(), GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	// Initialize positions and velocities, and update them into OpenGL buffers
	current_particle_count = desired_particle_count;
	reset_particles();

	// Create VAOs to render the particles. The positions are at location '0' and the colors at '1'.
	// Since there are two buffers with positions, we need two VAOs.
	glGenVertexArrays(2, particle_vao);
	glBindVertexArray(particle_vao[0]);
		glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[0]);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
		glBindBuffer(GL_ARRAY_BUFFER, particle_colors_bo);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glBindVertexArray(particle_vao[1]);
		glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[1]);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
		glBindBuffer(GL_ARRAY_BUFFER, particle_colors_bo);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	current_read = 0;
	current_write = 1;

	// Create the query object
	glGenQueries(1, &UpdateTimeQuery);
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	// Data of the main camera
	CameraData_ubo.SetProjection(
		glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 1.0f, 10000.0f));
	CameraData_ubo.SetCamera(the_camera);
	CameraData_ubo.UpdateOpenGLData();
	
	// Recreate particles if the number changed or when the type of computation changed
	if ((current_particle_count != desired_particle_count) ||
		(current_update_method != desired_update_method))
	{
		current_particle_count = desired_particle_count;
		current_update_method = desired_update_method;
		reset_particles();
	}

	//------------------------------------------------------------
	//--  Update the positions and velocities of all particles  --

	if (current_update_method == UPDATE_ON_CPU)
	{
		// Simulate the particles on CPU
		int start_time = glutGet(GLUT_ELAPSED_TIME);

		// "Rename" variables so that their name was the same as their name in shaders
		vector<glm::vec4> &particle_positions_read = particle_positions[current_read];
		vector<glm::vec4> &particle_positions_write = particle_positions[current_write];

		// Task 1: Implement n-body simulation on CPU: for each particle, compute the force to each other particle,
		//			then the acceleration, and then update the position and velocity.
		//		Hint: Variables:
		//			particle_positions_read   .. array of vec4, positions of particles to read from
		//			particle_positions_write  .. array of vec4, positions of particles to write into
		//			particle_velocities       .. array of vec4, velocities of particles to read from and to write into
		//			current_particle_count    .. integer, number of active particles in simulation
		//			distance_threshold        .. float, objects closer than this value exert no force to each other
		//			accel_factor              .. a constant that affects the final acceleration
		//			time_delta                .. time elapsed since the last frame of our simulation
		//		Hint: You may use most of GLSL functions and structures, but remember they are in glm namespace (like
		//			glm::vec3, glm::normalize, glm::dot, etc.). Unfortunately, in C++, there is no swizzling.

		// Update the position and velocity of each particle
		for (int i = 0; i < current_particle_count; i++)
		{
			// Read the initial position and velocity of the particle
			glm::vec3 position = glm::vec3(particle_positions_read[i]);
			glm::vec3 velocity = glm::vec3(particle_velocities[i]);

			// Compute the acceleration of the particle. We simulate the gravitation force between all particles.
			// We compute a vector from this particle to each other particle, and if it is further than some distance
			// threshold (to avoid division by zero), we compute a force that attracts the particle. The force
			// fades with the square of the distance, as in real world.
			glm::vec3 accel = glm::vec3(0.0f);
			for (int p = 0; p < current_particle_count; p++)
			{
				// Position of the other particle
				glm::vec3 other = glm::vec3(particle_positions_read[p]);
				// Vector between these two particles
				glm::vec3 direction = other - position;
				// Square of the distance between these particles
				float dist_sq = glm::dot(direction, direction);
				if (dist_sq > distance_threshold)
				{
					// Add the force to acceleration. We must normalize the 'diff' to get the direction,
					// and fade it with the square of the distance (division by 'dist_sq').
					accel += glm::normalize(direction) / dist_sq;
				}
			}
			// We multiply the acceleration with some constant. The constant contains all physics stuff like
			// the mass of particles, gravitational constants etc.
			accel *= accel_factor;

			// Update the position and the velocity
			position += velocity * time_delta + 0.5f * accel * time_delta * time_delta;
			velocity += accel * time_delta;

			// Store the result. Note that we must store the positions into a different buffer than we read it from,
			// because we need the initial values for simulation of other particles
			particle_velocities[i] = glm::vec4(velocity, 0.0f);
			particle_positions_write[i] = glm::vec4(position, 1.0f);
		}

		int end_time = glutGet(GLUT_ELAPSED_TIME);
		int elapsed_time_ms = end_time - start_time;
		update_time_ms = float(elapsed_time_ms);

		// Update the GPU buffer to render the positions.
		glBindBuffer(GL_ARRAY_BUFFER, particle_positions_bo[current_write]);
		glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(glm::vec4) * current_particle_count, particle_positions[current_write].data());
		glBindBuffer(GL_ARRAY_BUFFER, 0);
	}
	else if ((current_update_method == UPDATE_ON_GPU_SIMPLE) || (current_update_method == UPDATE_ON_GPU_SHARED))
	{
		// Simulate the particles on GPU using nbody_compute.glsl or nbody_shared_compute.glsl

		// Bind all buffers to OpenGL
		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, particle_positions_bo[current_read]);
		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, particle_positions_bo[current_write]);
		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, particle_velocities_bo);
		// Memory barrier, we need the data in the buffers to be ready
		glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

		// Choose the proper program
		if ((current_update_method == UPDATE_ON_GPU_SIMPLE) && update_particles_program.IsValid())
		{
			update_particles_program.Use();

			// Update uniform values
			update_particles_program.Uniform1i("current_particle_count", current_particle_count);
			update_particles_program.Uniform1f("time_delta", time_delta);
			update_particles_program.Uniform1f("accel_factor", accel_factor);
			update_particles_program.Uniform1f("distance_threshold", distance_threshold);

			const int local_size_x = 256;		// This must be the same as 'layout (local_size_x = 256) in;' in nbody_compute.glsl

			// Dispatch the compute shader
			glBeginQuery(GL_TIME_ELAPSED, UpdateTimeQuery);
			glDispatchCompute(current_particle_count / local_size_x, 1, 1);
			glEndQuery(GL_TIME_ELAPSED);
		}
		else if ((current_update_method == UPDATE_ON_GPU_SHARED) && update_particles_shared_program.IsValid())
		{
			update_particles_shared_program.Use();

			// Update uniform values
			update_particles_shared_program.Uniform1i("current_particle_count", current_particle_count);
			update_particles_shared_program.Uniform1f("time_delta", time_delta);
			update_particles_shared_program.Uniform1f("accel_factor", accel_factor);
			update_particles_shared_program.Uniform1f("distance_threshold", distance_threshold);

			const int local_size_x = 256;		// This must be the same as 'layout (local_size_x = 256) in;' in nbody_shared_compute.glsl

			// Dispatch the compute shader
			glBeginQuery(GL_TIME_ELAPSED, UpdateTimeQuery);
			glDispatchCompute(current_particle_count / local_size_x, 1, 1);
			glEndQuery(GL_TIME_ELAPSED);
		}
		else
		{
			// Without this, UpdateTimeQuery object would not be created, and debug layers of OpenGL would print error messages
			glBeginQuery(GL_TIME_ELAPSED, UpdateTimeQuery);
			glEndQuery(GL_TIME_ELAPSED);
		}

		// Evaluate the query
		glFinish();					// Wait for OpenGL, don't forget OpenGL is asynchronous
		GLuint64 render_time;
		glGetQueryObjectui64v(UpdateTimeQuery, GL_QUERY_RESULT, &render_time);
		update_time_ms = float(render_time) * 1e-6f;
	}
}

/// Renders the whole frame
void render_scene()
{
	// Render into the main window
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, win_width, win_height);

	// Clear the color buffer. Ignore the depth buffer, we do not use it
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	// Set the data of the camera
	CameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);

	// Set up very simple blending
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);

	// Choose proper shader to render the particles
	if ((!use_geometry_shaders && particle_simple_program.IsValid()) ||
		(use_geometry_shaders && particle_textured_program.IsValid()))
	{
		if (use_geometry_shaders)
		{
			// Set the necessary values
			particle_textured_program.Use();
			particle_textured_program.Uniform1f("particle_size_vs", particle_size);
			glActiveTexture(GL_TEXTURE0);		glBindTexture(GL_TEXTURE_2D, particle_tex);
		}
		else
		{
			// Set the necessary values
			particle_simple_program.Use();
			glPointSize(2.0f);
		}
		
		// Bind proper VAO (use the VAO with the data we just wrote)
		glBindVertexArray(particle_vao[current_write]);
		// Memory barrier, we need the data in the buffers to be ready
		glMemoryBarrier(GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT);
		// Draw the particles as points
		glDrawArrays(GL_POINTS, 0, current_particle_count);
	}

	// Disable blending
	glDisable(GL_BLEND);

	// Swap indices of read and write buffers with positions
	std::swap(current_read, current_write);

	//----------------------------------------------

	// Reset the VAO, the program, and the FBO
	glBindVertexArray(0);
	glUseProgram(0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void TW_CALL reset(void *)
{
	reset_particles();
}

void init_gui()
{
	// Initial values
	use_geometry_shaders = false;
	particle_size = 0.5f;
	
	reset_particles_seed = 1;
	accel_factor = 0.2f;
	distance_threshold = 0.01f;
	time_delta = 0.005f;

	desired_particle_count = 256;
	max_particle_count = 131072;

	current_update_method = desired_update_method = UPDATE_ON_CPU;

	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);

	TwAddVarRO(the_gui, "Update time (ms)", TW_TYPE_FLOAT, &update_time_ms, nullptr);
	
	TwAddVarRW(the_gui, "Use GS", TW_TYPE_BOOLCPP, &use_geometry_shaders, nullptr);
	TwAddVarRW(the_gui, "Particle size", TW_TYPE_FLOAT, &particle_size, "min=0.05 max=3.0 step=0.05");

	TwAddVarRW(the_gui, "Seed", TW_TYPE_INT32, &reset_particles_seed, "min=1 max=100 step=1");

	TwEnumVal particle_counts[10] = {
		{256, "256"},
		{512, "512"},
		{1024, "1024"},
		{2048, "2048"},
		{4096, "4096"},
		{8192, "8192"},
		{16384, "16384"},
		{32768, "32768"},
		{65536, "65536"},
		{131072, "131072"},
	};

	TwType countsType = TwDefineEnum("ParticleCounts", particle_counts, 10);
	TwAddVarRW(the_gui, "Particle Count", countsType, &desired_particle_count, nullptr);

	TwAddButton(the_gui, "Reset Particles", reset, nullptr, nullptr);

	TwEnumVal computation[3] = {
		{UPDATE_ON_CPU, "Update on CPU"},
		{UPDATE_ON_GPU_SIMPLE, "Update on GPU"},
		{UPDATE_ON_GPU_SHARED, "Update on GPU (shared)"},
	};

	TwType computationType = TwDefineEnum("Computations", computation, 3);
	TwAddVarRW(the_gui, "Computation", computationType, &desired_update_method, nullptr);

	TwAddVarRW(the_gui, "Acceleration Factor", TW_TYPE_FLOAT, &accel_factor, "min=0.01 max=1.0 step=0.01");
	TwAddVarRW(the_gui, "Distance Threshold", TW_TYPE_FLOAT, &distance_threshold, "min=0.01 max=1.0 step=0.01");
	TwAddVarRW(the_gui, "Time delta", TW_TYPE_FLOAT, &time_delta, "min=0.001 max=0.010 step=0.001");
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Lesson 7: Particle systems, compute shaders, geometry shaders");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}